from enum import Enum

from selve.protocol import MethodCall
from selve.protocol import ParameterType
from selve.protocol import DeviceType
from selve.protocol import CommandType
from selve.utils import singlemask
from selve.utils import true_in_list
from selve.utils import b64bytes_to_bitlist
import logging

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.DEBUG)

class CommeoCommand(Enum):
    GETINFO = "getInfo"
    GETIDS = "getIDs"
    GETVALUES = "getValues"
    RESULT = "result"
    DEVICE = "device"

class GatewayDeviceCommand(MethodCall):
    #For every gateway device command starting with selve.GW.device
    def __init__(self, method_name, parameters = []):
         super().__init__("selve.GW.device." + method_name.value, parameters)

class CommandGatewayGetIDs(GatewayDeviceCommand):
    #selve.GW.device.getIDs
    def __init__(self):
        super().__init__(CommeoCommand.GETIDS)

    def process_response(self, methodResponse):
        self.ids = [ b for b in true_in_list(b64bytes_to_bitlist(methodResponse.parameters[0][1]))]
        _LOGGER.debug(self.ids)

class CommandGatewayGetProperties(GatewayDeviceCommand):
    #selve.GW.device.getInfo
    #for CommeoDevice.discover_properties
    def __init__(self, commeoID):
        super().__init__(CommeoCommand.GETINFO, [(ParameterType.INT, commeoID)])

    def process_response(self, methodResponse):
        self.name = methodResponse.parameters[0][1]
        self.activity = methodResponse.parameters[2][1]
        self.deviceType = DeviceType(int(methodResponse.parameters[3][1]))

class CommandGatewayGetState(GatewayDeviceCommand):
    #selve.GW.device.getValues
    def __init__(self, commeoID):
        super().__init__(CommeoCommand.GETVALUES, [(ParameterType.INT, commeoID)])

    def process_response(self, methodResponse):
        self.position = methodResponse.parameters[3][1]
        self.status = methodResponse.parameters[2][1]

class CommandCommeo(MethodCall):
    # For every commeo command starting with selve.GW.command.
    def __init__(self, method_name, parameters = []):
         super().__init__("selve.GW.command." + method_name.value, parameters)

class CommandCommeoDevice(CommandCommeo):
    #selve.GW.command.device
    def __init__(self, commeoID, command, mode = 1, parameter = 0): #mode 1 = manuell, mode 2 = automatik
        super().__init__(CommeoCommand.DEVICE, [(ParameterType.INT, commeoID), (ParameterType.INT, command.value), (ParameterType.INT, mode), (ParameterType.INT, parameter)])

class State:
    def __init__(self, status = 0, position = 0):
        self.status = int(status)
        self.position = int(position)

    def setstate(self, status, position):
        self.position = int(position)
        self.status = int(status)

class CommeoDevice:
    def __init__(self, gateway, commeoID, discover=False):
        self.state = State()
        self.commeoID = commeoID
        self.gateway = gateway
        self.device_type = DeviceType.UNKNOWN
        self.name = "Not defined"
        if discover:
            self.discover_properties()
        self.updateState()

    def stop(self, automatic=False):
        self.executeCommand(CommandType.STOP, automatic)

    def moveDown(self, automatic=False):
        self.executeCommand(CommandType.DEPARTURE, automatic)

    def moveUp(self, automatic=False):
        self.executeCommand(CommandType.DRIVEAWAY, automatic)

    def moveIntermediatePosition1(self, automatic=False):
        self.executeCommand(CommandType.POSITION_1, automatic)

    def moveIntermediatePosition2(self, automatic=False):
        self.executeCommand(CommandType.POSITION_2, automatic)

    def moveToPosition(self, pos, automatic=False):
        pos = round(pos*655.35)
        self.executeCommand(CommandType.DrivePos, automatic, pos)

    def executeCommand(self, commandType, automatic=False, parameter = 0):
        if automatic:
            command = CommandCommeoDevice(self.commeoID, commandType, 2, parameter)
        else:
            command = CommandCommeoDevice(self.commeoID, commandType, 1, parameter)
        command.execute(self.gateway)
        return command

    def discover_properties(self):
        command = CommandGatewayGetProperties(self.commeoID)
        command.execute(self.gateway)
        self.device_type = command.deviceType
        self.name = command.name

    def updateState(self):
        command = CommandGatewayGetState(self.commeoID)
        command.execute(self.gateway)
        self.state.setstate(command.status, command.position)

    def __str__(self):
        return "Device of type: " + self.device_type.name + " on channel " + str(
            self.commeoID) + " with name " + self.name




from enum import Enum
from itertools import chain
import untangle
import logging


_LOGGER = logging.getLogger(__name__)
class DeviceType(Enum):
    UNKNOWN = 0
    SHUTTER = 1
    BLIND = 2
    AWNING = 3
    SWITCH = 4
    DIMMER = 5
    NIGHT_LIGHT = 6
    DRAWN_LIGHT = 7
    HEATING = 8
    COOLING = 9
    COOLING2 = 10
    GATEWAY = 11

class CommandType(Enum):
    STOP = 0
    DRIVEAWAY = 1 #driveUp
    DEPARTURE = 2 #drivedown
    POSITION_1 = 3
    POSITION_2 = 4
    DrivePos = 7

class ParameterType(Enum):
    INT = "int"
    STRING = "string"
    BASE64 = "base64"

class MethodCall:

    def __init__(self, method_name, parameters = []):
        self.method_name = method_name
        self.parameters = parameters

    def serializeToXML(self):
        xmlstr = "<methodCall>"
        xmlstr += "<methodName>"+self.method_name+"</methodName>"
        if (len(self.parameters) > 0):
            xmlstr += "<array>"
            for typ, val in self.parameters:
                xmlstr+="<{0}>{1}</{0}>".format(typ.value, val)
            xmlstr += "</array>"
        xmlstr+= "</methodCall>"
        return xmlstr.encode('utf-8')
    
    def execute(self, gateway):
        response = gateway.executeCommand(self)
        if response != None and isinstance(response, MethodResponse):
            self.process_response(response)
    
    def process_response(self, methodResponse):
        _LOGGER.debug(methodResponse)


class MethodResponse:

    def __init__(self, name, parameters):
        self.name = name
        self.parameters = parameters

class ErrorResponse:

    def __init__(self, message, code):
        self.message = message
        self.code = code


def create_error(obj):
    return ErrorResponse(obj.methodResponse.fault.array.string.cdata, obj.methodResponse.fault.array.int.cdata ) 

def create_response(obj):
    array = obj.methodResponse.array
    methodName = list(array.string)[0].cdata
    str_params_tmp = list(array.string)[1:]
    str_params = [(ParameterType.STRING, v.cdata) for v in str_params_tmp]
    int_params = []
    if hasattr(array, ParameterType.INT.value):
        int_params = [(ParameterType.INT, v.cdata) for v in list(array.int)]
    b64_params = []
    if hasattr(array, ParameterType.BASE64.value):
        b64_params = [(ParameterType.BASE64, v.cdata) for v in list(array.base64)]
    paramslist = [str_params, int_params, b64_params]
    flat_params_list = list(chain.from_iterable(paramslist))
    return MethodResponse(methodName, flat_params_list)


def process_response(xmlstr):
    _LOGGER.debug(str(xmlstr))
    #The selve device sometimes answers a badformed header. This is a patch
    xmlstr = str(xmlstr).replace('<?xml version="1.0"? encoding="UTF-8">', '<?xml version="1.0" encoding="UTF-8"?>')
    res = untangle.parse(xmlstr)
    if not hasattr(res, 'methodResponse'):
        _LOGGER.error("Bad response format")
        return None
    if hasattr(res.methodResponse, 'fault'):
        return create_error(res)
    return create_response(res)

def main():

    selve_string = u'''<?xml version="1.0"? encoding="UTF-8">
    <methodResponse>
        <array>
            <string>Methodenname</string>
            <string>Parameter 1</string>
            <int>Parameter 2</int>
            <base64>Parameter 3</base64>
        </array>
    </methodResponse>
    '''

    error_string = '''<?xml version="1.0" encoding="UTF-8"?>
    <methodResponse>
        <fault>
            <array>
                <string>Method not supported!</string>                
                <int>2</int>                
            </array>
        </fault>
    </methodResponse>
    '''

    a ='<?xml version="1.0" encoding="UTF-8"?>\r\n<methodResponse>\r\n\t<array>\r\n\t\t<string>selve.GW.iveo.getIDs</string>\r\n\t\t<base64>fx4AAAAAAAA=</base64>\r\n\t</array>\r\n</methodResponse>\r\n\n'

    process_response(selve_string)


if __name__ == '__main__':
    main()



#!/usr/bin/python

import time
import serial
from enum import Enum
import logging

from selve.utils import *
from selve.protocol import process_response
from selve.commeo import *

_LOGGER = logging.getLogger(__name__)

class Gateway():   

    def __init__(self, port, discover = True):
        """                
        Arguments:
            port {String} -- Serial port string as it is used in pyserial
        
        Keyword Arguments:
            discover {bool} -- True if the gateway should try to discover 
                               the devices on init (default: {True})
        """
        self.port = port
        if discover:
            self.discover()
    
    def configserial(self):
        """
        Configure the serial port
        """
        self.ser = serial.Serial(
            port=self.port,
            baudrate=115200,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            bytesize=serial.EIGHTBITS)
        self.ser.timeout = 0
        self.ser.xonxoff = False
        self.ser.rtscts = False
        self.ser.dsrdtr = False
        self.ser.writeTimeout = 2

            
    def executeCommand(self, command):
        """[summary]
        Execute the given command using the serial port.
        It opens a communication to the serial port each time a
        command is executed.
        At this moment it doesn't keep a queue of commands. If
        a command blocks the serial it will wait.
        
        Arguments:
            command {protocol.MethodCall} -- Command to be send 
            through the serial port
        
        Returns:
            MethodResponse -- if the command was executed 
            sucessufully
            ErrorResponse -- if the gateway returns an error
        """
        commandstr = command.serializeToXML()
        _LOGGER.info('Gateway writting: ' + str(commandstr))

        try:
            self.configserial()

        except Exception as e:            
            _LOGGER.error ('error open serial port: ' + str(e))
            exit()
        
        if self.ser.isOpen():
            try:
                self.ser.flushInput()
                self.ser.flushOutput()
                
                self.ser.write(commandstr)
                time.sleep(0.5)
                response_str = "" 
                while True:
                    response = self.ser.readline().strip()
                    response_str += response.decode()
                    if (response.decode() == ''):
                        break
                    
                self.ser.close()
                _LOGGER.info('read data: ' + response_str)
                return process_response(response_str)
            except Exception as e1:
                _LOGGER.exception ("error communicating...: " + str(e1))
        else:
            _LOGGER.error ("cannot open serial port")
        
        return None


    def discover(self):
        """[summary]
            Discover all devices registered on the usb-commeo        
        """
        command = CommandGatewayGetIDs()
        num_retries = 5
        retry_n = 0
        while not hasattr(command, "ids") and retry_n <=num_retries:
            command.execute(self)
            retry_n += 1
            time.sleep(1)
        
        if not hasattr(command, "ids"):
            _LOGGER.info("Associated Devices not found") 
            self.devices = {}
        else:
            _LOGGER.debug(f'discover ids: {command.ids}')
            self.devices = dict([(id, CommeoDevice(self, id , True) )for id in command.ids])
        self.list_devices() 
       

    def is_id_registered(self, id):
        """[summary]
        check if a device id is registered on the gateway
        Arguments:
            id {int} -- Device id to check
        
        Returns:
            boolean -- True if the id is registered
                       False otherwise
        """
        return id in self.devices
        
    def list_devices(self):
        """[summary]
        Print the list of registered devices
        """
        for id, device in self.devices.items():
            print(str(device))
            #_LOGGER.info(str(device))

if __name__ == '__main__':
    #print (singlemask(2).decode('utf-8'))
    _LOGGER.setLevel(logging.DEBUG)
    #manual = MethodCall("selve.GW.iveo.getIDs",[])
    portname = '/dev/cu.usbserial-DJ00T875'
    gat = Gateway(portname, False)
    gat.discover()
    devices = list(gat.devices.values())
    #device = IveoDevice(gat, 0)
    #device.stop(False)
    #gat.list_devices()

    # device1 = IveoDevice(gat, 1)
    # device2 = IveoDevice(gat, 2)
    # device1.moveDown()
    # time.sleep(5)
    # device1.moveUp()
    # device2.moveDown()
    # time.sleep(3)
    # device1.stop()
    # device2.stop
    # time.sleep(1)
    # device1.stop()
    #command = IveoCommandGetIds()
    #command.execute(gat)
    #print (command.ids)
    #response = b'<?xml version="1.0"? encoding="UTF-8">\r\n<methodResponse>\r\n\t<array>\r\n\t\t<string>selve.GW.iveo.commandManual</string>\r\n\t\t<int>1</int>\r\n\t</array>\r\n</methodResponse>\r\n\n<?xml version="1.0"? encoding="UTF-8">\r\n<methodCall>\r\n<methodName>selve.GW.iveo.commandResult</methodName>\r\n\t<array>\r\n\t\t<int>0</int>\r\n\t\t<base64>AQAAAAAAAAA=</base64>\r\n\t\t<int>0</int>\r\n\t</array>\r\n</methodCall>\r\n\n'
    #c = response.decode()
    #print(c)
    #deserialize(response)

